/*
 * @(#)ChargedFluid.java		1.10 07/01/20
 *
 * ChargedFluid package
 *
 * COPYRIGHT NOTICE
 * Copyright (c) 2007 Herbert H.H. Chang, Daniel J. Valentino, Gary R. Duckwiler, and Arthur W. Toga
 * Laboratory of Neuro Imaging, Department of Neurology, UCLA.
 */
 
package tools;

import java.awt.Point;
import java.awt.Polygon;
import java.awt.geom.Point2D;
import java.util.LinkedList;
import java.util.Iterator;

/**
 * The <code>ChargedFluid</code> class contains charges that are stores in the
 * <code>LinkedList</code>. The charges in the <code>LinkedList</code> are not in position
 * order and there is no way to know the spacial relationship among the charges during the
 * evolution process. Once the evolution is over, a contour tracing technique is used to
 * reorganize the sequence of the charges in the <code>LinkedList</code> such that the
 * positions of them can be used to create a <code>Polygon</code>. The surrounded area (pixels)
 * by this contour can then be obtained using the <code>Polygon</code> class.
 */
public class ChargedFluid
{
	/** 
	 * Capturing the surrounded area by the contour. 
	 */
	public Polygon polygonCF;
	
	/** 
	 * Storing the fluid elements (Charge class). 
	 */
	public LinkedList linkedList;
		
	/** 
	 * The total charge after normalization. 
	 */
	public double totalCharge;
	
	/** 
	 * True if the Charged Fluid is frozen. 
	 */
	public boolean isFrozen;
	
	/** 
	 * The number of traced charges at the previous step.
	 * Used for the stop criteria of the evolution.
	 */
	public int previousNTraceCharge;
	
	/** 
	 * The number of traced charges at the current step.
	 * Used for the stop criteria of the evolution.
	 */
	public int currentNTraceCharge;
	
	/** 
	 * The minimum position of all charges in x-axis.
	 */
	public int minXAxis;
	
	/** 
	 * The maximum position of all charges in x-axis. 
	 */
	public int maxXAxis;
	
	/** 
	 * The minimum position of all charges in y-axis. 
	 */
	public int minYAxis;
	
	/** 
	 * The maximum position of all charges in y-axis. 
	 */
	public int maxYAxis;
	
		
	/** 
	 * Constructs a charged fluid with the specified linked list.
	 * @param	linkedList	Storing the charges.
	 */
	public ChargedFluid(LinkedList linkedList)
	{
		this.linkedList = linkedList;
	}
	
	
	/** 
	 * Returns the width (x-axis) of the charged fluid.
	 * @return	the width of the charged fluid.
	 */
	public int getWidth()
	{
		return (maxXAxis - minXAxis + 1);
	}
	
	
	/** 
	 * Returns the height (y-axis) of the charged fluid. 
	 * @return	the height of the charged fluid.
	 */
	public int getHeight()
	{
		return (maxYAxis - minYAxis + 1);
	}
	
	
	/** 
	 * Returns the center point of the Charged Fluid based on the rectangle consisted by
	 * <tt>minXAxis</tt>, <tt>maxXAxis</tt>, <tt>minYAxis</tt> and <tt>maxYAxis</tt>.
	 * @return	the point of the charged fluid center.
	 */
	public Point getCenter()
	{
		return new Point((maxXAxis+minXAxis)/2 , (maxYAxis+minYAxis)/2);
	}
	
	
	/** 
	 * Computes the contour polygon based upon the charges in the linked list. The polygon
	 * represents the contour of the charged fluid.
	 */
	public void computePolygon()
	{
		int contourLength = linkedList.size();
		// Create the arrays for storing the charge positions.
		int[] xPoints = new int[contourLength];
		int[] yPoints = new int[contourLength];
		for (int i = 0; i < contourLength; i++) {
			Charge currentCharge = (Charge)linkedList.get(i);
			Point chargePt = currentCharge.position;
			xPoints[i] = chargePt.x;
			yPoints[i] = chargePt.y;
		}
		// Create a polygon.
		polygonCF = new Polygon(xPoints, yPoints, contourLength);
	}
}