/*
 * @(#)Matrix.java		1.10 07/01/20
 *
 * ChargedFluid package
 *
 * COPYRIGHT NOTICE
 * Copyright (c) 2007 Herbert H.H. Chang, Daniel J. Valentino, Gary R. Duckwiler, and Arthur W. Toga
 * Laboratory of Neuro Imaging, Department of Neurology, UCLA.
 */

package tools;

import java.awt.Point;
import java.awt.Color;
import java.awt.Rectangle;
import java.awt.Polygon;
import java.lang.Math;
import java.util.Vector;
import java.util.LinkedList;
import java.util.Iterator;
import java.awt.image.BufferedImage;
import java.awt.Graphics2D;
import java.awt.AlphaComposite;

/** 
 * The <code>Matrix</code> class is dedicated to the computation and display related to 
 * the original image intensity and gradient. It also handles the generation of the intial
 * contours and the ouput of the final results in different formats.
 */
public class Matrix
{
	/**
	 * Buffered image of the input data.
	 */
	private BufferedImage _bufferedImage;

	/**
	 * The copy of the original image RGB data array. 
	 */
	private int[][] _reservedImageRGBArray;
	
	/**
	 * Image intensity array. 
	 */
	private double[][] _imageIntensityArray;
	
	/**
	 * Image gradient array subject to the modification of the saturation function. 
	 */
	private double[][] _imageGradientArray;
	
	/**
	 * Image gradient array in x-axis. 
	 */
	private double[][] _gradientArrayInX;
	
	/**
	 * Image gradient array in y-axis. 
	 */
	private double[][] _gradientArrayInY;
	
	/**
	 * Original image gradient array without any modification. 
	 */
	private double[][] _reservedGradientArray;
	
	/** 
  	 * Boolean array for indicating the free charged fluid trace. True, if the position is 
  	 * passed by the free charged fluid. T:(+) passed and F:(-) empty. The update of this
  	 * array is performed for each individual fluid element right after it deforms the shape.
  	 */
  	private boolean[][] _traceBooleanArray;
  	
  	/** 
  	 * Boolean array for indicating the free charged fluid trace. True, if the position is 
  	 * passed by the free charged fluid. T:(+) passed and F:(-) empty. The update of this
  	 * array is performed after all charged fluid elements have deformed their shape.
  	 */
  	private boolean[][] _labelBooleanArray;
  	
  	/**
	 * Representing the final areas inside the specified contour with true and 
	 * false for outside.
	 */
	public boolean[][] booleanArray;
	
	/**
	 * Storing the <code>ChargedFluid</code> class. 
	 */
	private Vector _chargedFluidVector = new Vector();
	
	/**
	 * Storing the <code>Poisson</code> class. 
	 */
	private Vector _poissonVector = new Vector();
	
	/**
	 * Storing the indexes (<code>ChargedFluidIndex</code>) of connected charged fluids
	 * to the charged fluid in the array. 
	 */
	private LinkedList[] _connectedCFIndexLinkArray;
	
	/**
	 * Storing the computation time.
	 */
	public String timeStrg;
	
	/**
	 * Image height of the input data. 
	 */
	private int _imageHeight;
	
	/** 
	 * Image width of the input data. 
	 */
	private int _imageWidth;
	
	/** 
	 * Initial box height to constrain the charged fluid. 
	 */
	private int _initialBoxHeight = 16;
	
	/**
	 * Initial box width to constrain the charged fluid. 
	 */
	private int _initialBoxWidth = 16;
	
	/** 
	 * Indicator to pause the program. 
	 */
	private int _pauseInteger;
	
	/**
	 * Index of using what kind of filters. 
	 */
	private int _indexOfFilter;
	
	/**
	 * Index of using what kind of computation to enhance the image gradient. 
	 */
	private int _indexOfPower;
	
	/** 
	 * The initial ensemble charge for each charged fluid. 
	 */
	private double _totalCharge = 5000.0;
	
	/**
	 * The normalized electric potential through the evolution process. 
	 */
	private double _normalizedPotential = 10000.0;
	
	/**
	 * The contour color of the charged fluid during the evolution. 
	 */
	private Color _contourColor = Color.yellow;
	
	/**
	 * The contour color (red) of the frozen charged fluid after the process. 
	 */
	private final int _shownParticleRed = 0xffff0000;
	
	/** 
  	 * Weighting factor for the electric potential to balance the image potential, which 
  	 * is the modulus of the image gradient. 
  	 */
	private double _beta;
	
	/**
	 * The specified condition for achieving the electrostatic equilibrium for the system. 
	 */
	private double _gamma;
	
	/**
	 * True, if the result is superimposed by the input image for display.
	 */
	public boolean isSuperimposed;
	
	/**
	 * True, if shows the original intensity image as background. 
	 */
	public boolean isIntensityImage;
	
	/**
	 * True, if the geometry of the initial contour is square. 
	 */
	public boolean isGammaNormal;
	
	/**
	 * True, if the system is terminated by users. 
	 */
	private boolean _isBreak;
	
	/**
	 * True, if the geometry of the initial contour is square. 
	 */
	private boolean _isSquare = true;
	
	/**
	 * True, if the image gradient is directly used as the effective field.
	 */
	public boolean isGradientField;
	
	/**
	 * Storing the seed points (<code>Point</code>) of the charged fluids. 
	 */
	private Vector _seedPoint = new Vector();
	
	/**
	 * Storing the <code>ChargedFluid</code> class for display purpose. 
	 */
	private Vector _tempCFVector = new Vector(3);
	

	/**
	 * Constructs a matrix with the specified integer and double parameters.
	 * @param	intParameter	integer parameters that contain	the index of the computation
	 *							power and the index of the filter type.
	 * @param	parameter		double parameter that contain the value of beta.
	 * @param	booleanPar		boolean values indicate if the contours and results are shown 
	 *							with the original image data, if displaying the intensity image
	 *							, if the gamma value is normal, and if the image gradient
	 *							is directly used as the effective field.
	 */
	public Matrix(int[] intParameter, double[] parameter, boolean[] booleanPar)
	{
		_indexOfPower = intParameter[0];
		_indexOfFilter = intParameter[1];
		
		_beta = parameter[0];
		
		isSuperimposed = booleanPar[0];
		isIntensityImage = booleanPar[1];
		isGammaNormal = booleanPar[2];
		isGradientField = booleanPar[3];
		
		if (isGammaNormal) {
			_gamma = 0.03;
		}
		else {
			_gamma = 0.01;
		}
	}
  
  
	/**
	 * Calculates the computation array size and allocates the arrays with the specified
	 * buffered image and the raw image data array.
	 * @param	bufferedImage	the buffered image from the original image data.
	 * @param	imageArray		the raw intensity array from the original image data.
	 */
	public void initialization(BufferedImage bufferedImage, double[][] imageArray)
	{
		_imageHeight = imageArray.length;
		_imageWidth = imageArray[0].length;
		_bufferedImage = bufferedImage;
		_reservedImageRGBArray = new int[_imageHeight][_imageWidth];
		
		for (int y = 0; y < _imageHeight; y++) {
			for (int x = 0; x < _imageWidth; x++) {
				_reservedImageRGBArray[y][x] = bufferedImage.getRGB(x, y);
			}
		}
		
		_traceBooleanArray = new boolean[_imageHeight][_imageWidth];
		_labelBooleanArray = new boolean[_imageHeight][_imageWidth];
		_reservedGradientArray = new double[_imageHeight][_imageWidth];
		_imageGradientArray = new double[_imageHeight][_imageWidth];
		_gradientArrayInX = new double[_imageHeight][_imageWidth];
		_gradientArrayInY = new double[_imageHeight][_imageWidth];
		_imageIntensityArray = imageArray;

		_computeGaussianFilterArray();
	}
	
  
  	/** 
  	 * Creates a new charged fluid with the input geometry contour.
  	 * @param	geometryContour		a contour consisting of charge points in square or
  	 * 								circle-like geometry.
  	 */
	public void createChargedFluid(GeometryContour geometryContour)
	{
		int numberOfElement = geometryContour.pointLinkedList.size();
		double charge = _totalCharge / (double)numberOfElement;
		LinkedList chargedFluidLink = new LinkedList();
		Iterator it = geometryContour.pointLinkedList.iterator();
	  	while (it.hasNext()) {
	    	Point chargePt = (Point)it.next();
			chargedFluidLink.add(new Charge(charge, chargePt));
	    }
		ChargedFluid newCF = new ChargedFluid(chargedFluidLink);
		newCF.totalCharge = _totalCharge;
		_chargedFluidVector.add(newCF);
		_setInitialRGBColor(chargedFluidLink);

		if (geometryContour.rectangle != null) {
			int xStart = geometryContour.rectangle.x;
			int yStart = geometryContour.rectangle.y;
			int xEnd = xStart + geometryContour.rectangle.width;
			int yEnd = yStart + geometryContour.rectangle.height;
			for (int y = yStart; y < yEnd; y++) {
				for (int x = xStart; x < xEnd; x++) {
					_traceBooleanArray[y][x] = true;
					_labelBooleanArray[y][x] = true;
				}
			}
		}
		else if (geometryContour.circlePolygon != null) {
			int[] xPoints = geometryContour.circlePolygon.xpoints;
			int[] yPoints = geometryContour.circlePolygon.ypoints;
			int nPoints = geometryContour.circlePolygon.npoints;
			for (int i = 0; i < nPoints; i++) {
				_traceBooleanArray[yPoints[i]][xPoints[i]] = true;
				_labelBooleanArray[yPoints[i]][xPoints[i]] = true;
			}
			Rectangle bounds = geometryContour.circlePolygon.getBounds();
			int xStart = bounds.x;
			int yStart = bounds.y;
			int xEnd = xStart + bounds.width;
			int yEnd = yStart + bounds.height;
			for (int y = yStart; y < yEnd; y++) {
				for (int x = xStart; x < xEnd; x++) {
					if (geometryContour.circlePolygon.contains(x,y)) {
						_traceBooleanArray[y][x] = true;
						_labelBooleanArray[y][x] = true;
					}
				}
			}
		}
		else {
			System.out.print("*** Geometry contour construct error ***");
		}
	}
		
  
	/** 
	 * Calculates the boundaries of Poissons. 
	 */
	public void computePoissonBound()
	{
		_computeChargedFluidSize();
		_constructPoisson();
	}
  
  		
  	/** 
  	 * Calculates the area of the boolean array from the ChargedFluids. Sets 
  	 * <code>true</code> to the inside of the polygon represeting the area.
  	 */
	public void computeBooleanArray()
	{
		booleanArray = new boolean[_imageHeight][_imageWidth];
		Iterator it = _chargedFluidVector.iterator();
		while (it.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)it.next();
		  	Polygon polygon = currentCF.polygonCF;
		  	for (int y = 0; y < _imageHeight; y++) {
				for (int x = 0; x < _imageWidth; x++) {
					if (polygon.contains(x, y)) {
	        			booleanArray[y][x] = true;
					}
				}
			}
	      	Iterator itc = currentCF.linkedList.iterator();
	      	while (itc.hasNext()) {
	        	Charge currentCharge = (Charge)itc.next();
	        	Point chargePt = currentCharge.position;
	        	booleanArray[chargePt.y][chargePt.x] = true;
	        }			
		}
	}
  	
  
	/** 
	 * Calculates the boundary of the free charged fluid in <tt>_chargedFluidVector</tt>. 
	 */
	private void _computeChargedFluidSize()
	{
		Iterator itv = _chargedFluidVector.iterator();
		while (itv.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)itv.next();
		  	Charge firstCharge = (Charge)currentCF.linkedList.getFirst();
		  	Point firstPt = firstCharge.position;
		  	int minXAxis = firstPt.x;
		  	int maxXAxis = firstPt.x;
		  	int minYAxis = firstPt.y;
		  	int maxYAxis = firstPt.y;
		  	for (int index = 1; index < currentCF.linkedList.size(); index++) {
		    	Charge currentCharge = (Charge)currentCF.linkedList.get(index);
		    	Point chargePt = currentCharge.position;
		    	if (chargePt.x > maxXAxis) {
		    		maxXAxis = chargePt.x;
		    	}
		    	if (chargePt.x < minXAxis) {
		    		minXAxis = chargePt.x;
		    	}
		    	if (chargePt.y > maxYAxis) {
		    		maxYAxis = chargePt.y;
		    	}
		    	if (chargePt.y < minYAxis) {
		    		minYAxis = chargePt.y;
		    	}
		    }
		    currentCF.maxXAxis = maxXAxis;
		    currentCF.minXAxis = minXAxis;
		    currentCF.maxYAxis = maxYAxis;
		    currentCF.minYAxis = minYAxis;
		}
	}
  
  
	/** 
	 * Constructs poissons based upon the bounds of the charged fluids.
	 */
	private void _constructPoisson()
	{
		_poissonVector.clear();
		int numberOfCF = _chargedFluidVector.size();
		_connectedCFIndexLinkArray = new LinkedList[numberOfCF];
		LinkedList indexCFLinkedList = new LinkedList();
		for (int k = 0; k < numberOfCF; k++) {
			indexCFLinkedList.add(new ChargedFluidIndex(k));
		}
		for (int i = 0; i < numberOfCF; i++) {
		  	ChargedFluid currentCF = (ChargedFluid)_chargedFluidVector.get(i);
		  	Point centerOfCF = currentCF.getCenter();
		  	int minXOfCF = currentCF.minXAxis;
		  	int minYOfCF = currentCF.minYAxis;
		  	int widthOfCF = currentCF.getWidth();
		  	int heightOfCF = currentCF.getHeight();
		  	Rectangle currentCFRect = new Rectangle(minXOfCF, minYOfCF, widthOfCF, heightOfCF);
		  	_connectedCFIndexLinkArray[i] = new LinkedList();
		  	_connectedCFIndexLinkArray[i].add(indexCFLinkedList.get(i));
		  	for (int j = i+1; j < numberOfCF; j++) {
		  		ChargedFluid candidateCF = (ChargedFluid)_chargedFluidVector.get(j);
		  		Point centerOfCand = candidateCF.getCenter();
		  		int minXOfCand = candidateCF.minXAxis;
		  		int minYOfCand = candidateCF.minYAxis;
		  		int widthOfCand = candidateCF.getWidth();
		  		int heightOfCand = candidateCF.getHeight();
		  		Rectangle candidateCFRect = new Rectangle(minXOfCand, minYOfCand, widthOfCand, heightOfCand);
		  		double realDistance = (centerOfCF.x - centerOfCand.x) * (centerOfCF.x - centerOfCand.x)
		  				+ (centerOfCF.y - centerOfCand.y) * (centerOfCF.y - centerOfCand.y);
		  		double kissDistance = _getThresholdDistance(currentCFRect, candidateCFRect);
		  		if (realDistance <= kissDistance) {
		  			_connectedCFIndexLinkArray[i].add(indexCFLinkedList.get(j));
		  		}
		  	}
		}
	
		for (int index = 0; index < numberOfCF; index++) {
			ChargedFluidIndex currentIndex = (ChargedFluidIndex)_connectedCFIndexLinkArray[index].getFirst();
			if (currentIndex.isNotIntegrated()) {
				LinkedList chargedFluidLinkedList = new LinkedList();
		    	Poisson newPoisson = new Poisson(chargedFluidLinkedList, _traceBooleanArray, _labelBooleanArray,
		    	 	_imageIntensityArray, _imageGradientArray, _gradientArrayInX, _gradientArrayInY, 
		    		_beta, _normalizedPotential, isGradientField);
		    	if (_connectedCFIndexLinkArray[index].size() > 1) {
		    		LinkedList integratedCF = new LinkedList();
		    		integratedCF.add(currentIndex);
		    		_uniteConnectedCFIndex(integratedCF, currentIndex.getIndex());
		    		ChargedFluid firstCF = (ChargedFluid)_chargedFluidVector.get(currentIndex.getIndex());
		    		chargedFluidLinkedList.add(firstCF);
		    		int minXAxis = firstCF.minXAxis;
		    		int maxXAxis = firstCF.maxXAxis;
		    		int minYAxis = firstCF.minYAxis;
		    		int maxYAxis = firstCF.maxYAxis;
		    		for (int i = 1; i < integratedCF.size(); i++) {
		    			ChargedFluidIndex indexCF = (ChargedFluidIndex)integratedCF.get(i);
		    			ChargedFluid currentCF = (ChargedFluid)_chargedFluidVector.get(indexCF.getIndex());
		    			_transferChargeBetweenChargedFluid(firstCF, currentCF);
		    			int currentMinXAxis = currentCF.minXAxis;
		    			int currentMaxXAxis = currentCF.maxXAxis;
		    			int currentMinYAxis = currentCF.minYAxis;
		    			int currentMaxYAxis = currentCF.maxYAxis;
		    			if (currentMinXAxis < minXAxis) {
		    				minXAxis = currentMinXAxis;
		    			}
		    			if (currentMaxXAxis > maxXAxis) {
		    				maxXAxis = currentMaxXAxis;
		    			}
		    			if (currentMinYAxis < minYAxis) {
		    				minYAxis = currentMinYAxis;
		    			}
		    			if (currentMaxYAxis > maxYAxis) {
		    				maxYAxis = currentMaxYAxis;
		    			}
		    		}
		    		newPoisson.minXAxis = minXAxis;
		    		newPoisson.maxXAxis = maxXAxis;
		    		newPoisson.minYAxis = minYAxis;
		    		newPoisson.maxYAxis = maxYAxis;
		    	}
		    	else {
		    		ChargedFluid currentCF = (ChargedFluid)_chargedFluidVector.get(currentIndex.getIndex());
		    		chargedFluidLinkedList.add(currentCF);
		    		newPoisson.minXAxis = currentCF.minXAxis;
		    		newPoisson.maxXAxis = currentCF.maxXAxis;
		    		newPoisson.minYAxis = currentCF.minYAxis;
		    		newPoisson.maxYAxis = currentCF.maxYAxis;
		    	}
		    	_poissonVector.add(newPoisson);
			}
		}
	}


	/** 
	 * Returns the threshold distance of two charged fluids with the specified bounds in
	 * rectangles <tt>CF1</tt> and <tt>CF2</tt>.
	 * @param	CF1		a rectangle storing the bounds of the first charged fluid.
	 * @param	CF2		a rectangle storing the bounds of the seond charged fluid.
	 * @return	the threshold distance between the given charged fluids.
	 */
	private double _getThresholdDistance(Rectangle CF1, Rectangle CF2)
	{
		int minXCF1 = CF1.x;
		int minYCF1 = CF1.y;
		int maxXCF1 = minXCF1 + CF1.width - 1;
		int maxYCF1 = minYCF1 + CF1.height - 1;
		int centXCF1 = (minXCF1 + maxXCF1) / 2;
		int centYCF1 = (minYCF1 + maxYCF1) / 2;
		int minXCF2 = CF2.x;
		int minYCF2 = CF2.y;
		int maxXCF2 = minXCF2 + CF2.width - 1;
		int maxYCF2 = minYCF2 + CF2.height - 1;
		int centXCF2 = (minXCF2 + maxXCF2) / 2;
		int centYCF2 = (minYCF2 + maxYCF2) / 2;
		double centDistInX = (double)(centXCF1 - centXCF2);
		double centDistInY = (double)(centYCF1 - centYCF2);
		if ((minXCF1 >= maxXCF2) || (maxXCF1 <= minXCF2)) {
			double kissDistance = (CF1.width + CF2.width) * (CF1.width + CF2.width) / 4.0
								+ (centDistInY * centDistInY);
			return kissDistance;
		}
		else if ((minYCF1 >= maxYCF2) || (maxYCF1 <= minYCF2)) {
			double kissDistance = (CF1.height + CF2.height) * (CF1.height + CF2.height) / 4.0
								+ (centDistInX * centDistInX);
			return kissDistance;
		}
		else {
			double kissDistance = (CF1.width + CF2.width) * (CF1.width + CF2.width) / 4.0
							+ (CF1.height + CF2.height) * (CF1.height + CF2.height) / 4.0;
			return kissDistance;
		}
	}
	

	/** 
	 * Unites the indexes of the charged fluids that are in the same <code>Poisson</code> 
	 * with the specified integrated list and the index of the target charged fluid.
	 * @param	integratedCF	a linked list that stores the indexes of the integrated 
	 *							charged fluids with the target charged fluid.
	 * @param	indexOfCF		the index of the target charged fluid.
	 */
	private void _uniteConnectedCFIndex(LinkedList integratedCF, int indexOfCF)
	{
		int connectedNumber = _connectedCFIndexLinkArray[indexOfCF].size();
		if (connectedNumber > 1) {
			for (int i = 1; i < connectedNumber; i++) {
				ChargedFluidIndex currentIndex = (ChargedFluidIndex)_connectedCFIndexLinkArray[indexOfCF].get(i);
				int index = currentIndex.getIndex();
				boolean isNewIndex = true;
				Iterator it = integratedCF.iterator();
				while (it.hasNext()) {
					ChargedFluidIndex indexGroup = (ChargedFluidIndex)it.next();
					if (index == indexGroup.getIndex()) {
						isNewIndex = false;
						break;
					}
				}
				if (isNewIndex) {
					integratedCF.add(currentIndex);
					currentIndex.setIndexIsNotIntegrated(false);
				}
				_uniteConnectedCFIndex(integratedCF, index);
			}
		}
	}

	
	/** 
	 * Transfers the charges from the specified second charged fluid to the specified 
	 * first charged fluid.
	 * @param	firstCF		the specified target charged fluid to accept the transferred
	 *						charges.
	 * @param	secondCF	the specified source charged fluid.
	 */
	private void _transferChargeBetweenChargedFluid(ChargedFluid firstCF, ChargedFluid secondCF)
	{
	  	Iterator itc = secondCF.linkedList.iterator();
	  	while (itc.hasNext()) {
	    	Charge currentCharge = (Charge)itc.next();
	    	firstCF.linkedList.add(currentCharge);
	    }
	}


	/** 
	 * Sets and refers the charged fluids from the <code>Poisson</code> domain back to the 
	 * <tt>_chargedFluidVector</tt> in the <code>Matrix</code> domain.
	 */
	private void _setChargedFluid2Vector()
	{
		_chargedFluidVector.clear();
	  	Iterator it = _poissonVector.iterator();
	  	while (it.hasNext()) {
	  		Poisson currentPn = (Poisson)it.next();
	      	Iterator itc = currentPn.linkedList.iterator();
	      	while (itc.hasNext()) {
	  			ChargedFluid currentCF = (ChargedFluid)itc.next();
	  			_chargedFluidVector.add(currentCF);
	  		}
	    }
	}

		
	/** 
	 * Calculates the initial parameters (electric field) for advancing the free 
	 * charged fluid at the very beginning. 
	 */
	public void computeInitialParameter()
	{
		_computeChargedFluidSize();
		_constructPoisson();
		Iterator itv = _poissonVector.iterator();
		while (itv.hasNext()) {
		  	Poisson currentPn = (Poisson)itv.next();
		  	currentPn.initialization();
		  	currentPn.computeInitialElectricField();
		}
	}
	

	/** 
	 * Updates the position, status and property of the free charged fluid for each
	 * iteration through the <code>Poisson</code>. 
	 */
	public void updateChargedFluid()
	{
		Iterator it = _poissonVector.iterator();
		while (it.hasNext()) {
		  	Poisson currentPn = (Poisson)it.next();
		  	if (!currentPn.isPoissonFrozen()) {
			  	currentPn.frontDeformation();
			  	currentPn.chargeDistribution(_gamma);
			  	currentPn.computeChargedFluidFront();
			  	currentPn.normalizeChargedFluid();
			  	currentPn.detectChargedFluidFrozen();
		  	}
		}
		_setChargedFluid2Vector();
		_showChargedFluidContour();
		_computeChargedFluidSize();
		_constructPoisson();
		Iterator itv = _poissonVector.iterator();
		while (itv.hasNext()) {
		  	Poisson currentPn = (Poisson)itv.next();
		  	currentPn.initialization();
		}
	}
  	
  	
  	/**
  	 * Adds the current CF to the tempCF vector for display purpose.
  	 */
  	public void addCF2Vector()
  	{
		Iterator itv = _chargedFluidVector.iterator();
		while (itv.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)itv.next();
		  	LinkedList chargedFluidLink = new LinkedList();
	      	Iterator itc = currentCF.linkedList.iterator();
	      	while (itc.hasNext()) {
	        	Charge currentCharge = (Charge)itc.next();
	        	chargedFluidLink.add(currentCharge);
	        }
		  	_tempCFVector.add(new ChargedFluid(chargedFluidLink));
		}
  	}
  	
  	
  	/** 
  	 * Determines whether all poissons are frozen or not.
  	 * @return	<code>true</code>, if all poissons are frozen.
  	 */
  	public boolean isAllPoissonFrozen()
  	{
		Iterator it = _poissonVector.iterator();
		while (it.hasNext()) {
		  	Poisson currentPn = (Poisson)it.next();
		  	if (!currentPn.isPoissonFrozen()) {
		  		return false;
		  	}
		}
		return true;
  	}


	/** 
	 * Recovers the buffered image to the original image for display. 
	 */
	public void recoverImage()
	{
	    if (isSuperimposed) {
	    	for (int y = 0; y < _imageHeight; y++) {
	      		for (int x = 0; x < _imageWidth; x++) {
					if (isIntensityImage) {
						_bufferedImage.setRGB(x, y, _reservedImageRGBArray[y][x]);
					}
					else {
						int gradientRGB = (0xff000000) | 
								((int)(255*_imageGradientArray[y][x]) << 16) |
                     			((int)(255*_imageGradientArray[y][x]) << 8) |
                     			((int)(255*_imageGradientArray[y][x]));
                    	_bufferedImage.setRGB(x, y, gradientRGB);		
					}	
	      		}	
			}
		}
		else {
			for (int y = 0; y < _imageHeight; y++) {
	      		for (int x = 0; x < _imageWidth; x++) {
	      			_bufferedImage.setRGB(x, y, 0);	
	      		}	
			}
		}
	}


	/** 
	 * Sets the initial RGB color to the buffered image <tt>_bufferedImage</tt> at the 
	 * charge positions in the specified charged fluid list for display.
	 * @param	chargedFluidLink	a new charged fluid linked list storing charges.
	 */
	private void _setInitialRGBColor(LinkedList chargedFluidLink)
	{
		Iterator it = chargedFluidLink.iterator();
		while (it.hasNext()) {
			Charge currentCharge = (Charge)it.next();
			Point pt = currentCharge.position;
			_bufferedImage.setRGB(pt.x, pt.y, _contourColor.getRGB());
		}
	}


	/** 
	 * Sets the RGB color to the buffered image <tt>_bufferedImage</tt> at the charge
	 * positions in the charged fluids for display.
	 */
	public void showCFEvolutionContours()
	{
		recoverImage();
		Iterator itv = _tempCFVector.iterator();
		while (itv.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)itv.next();
	      	Iterator it = currentCF.linkedList.iterator();
	      	while (it.hasNext()) {
	        	Charge currentCharge = (Charge)it.next();
	        	int x = currentCharge.position.x;
	        	int y = currentCharge.position.y;
	        	_bufferedImage.setRGB(x, y, _shownParticleRed);
	        }
		}
	}


	/** 
	 * Sets the RGB color to the buffered image <tt>_bufferedImage</tt> at the charge
	 * positions in the charged fluids for display.
	 */
	private void _showChargedFluidContour()
	{
		recoverImage();
		Iterator itv = _chargedFluidVector.iterator();
		while (itv.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)itv.next();
		  	if (currentCF.isFrozen) {
		      	Iterator it = currentCF.linkedList.iterator();
		      	while (it.hasNext()) {
		        	Charge currentCharge = (Charge)it.next();
		        	int x = currentCharge.position.x;
		        	int y = currentCharge.position.y;
		        	_bufferedImage.setRGB(x, y, _shownParticleRed);
		        }
		  	}
		  	else {
		      	Iterator it = currentCF.linkedList.iterator();
		      	while (it.hasNext()) {
		        	Charge currentCharge = (Charge)it.next();
		        	int x = currentCharge.position.x;
		        	int y = currentCharge.position.y;
		        	_bufferedImage.setRGB(x, y, _contourColor.getRGB());
		        }
			}
		}
	}
  	
  
	/** 
	 * Commputes the charged fluid front using a standard contour tracing algorithm to connect
	 * the charged fluid elements.
	 */
	public void traceChargedFluidFront()
	{
		Iterator itp = _poissonVector.iterator();
		while (itp.hasNext()) {
		  	Poisson currentPn = (Poisson)itp.next();
		  	currentPn.traceChargedFluidFront();
		}
		_setChargedFluid2Vector();
	}


	/** 
	 * Returns the final charged fluid contours displayed by polygons.
	 * @return	a <code>Vector</code> containing <code>Polygon</code> representing the charged
	 *			fluid contours. 
	 */
	public Vector getChargedFluidContour()
	{
		Vector polygonVector = new Vector();
		recoverImage();
		Iterator itp = _poissonVector.iterator();
		while (itp.hasNext()) {
		  	Poisson currentPn = (Poisson)itp.next();
			Iterator it = currentPn.linkedList.iterator();
			while (it.hasNext()) {
		  		ChargedFluid currentCF = (ChargedFluid)it.next();
		  		polygonVector.add(currentCF.polygonCF);
			}
		}
		return polygonVector;
	}
	
	
	/** 
  	 * Calculates the area of the boolean array w.r.t. the given point. Sets <code>true</code>
  	 * to the inside of the polygon on the array, which represents the area of the charged fluid.
  	 * It returns <code>true</code> if a charged fluid including the point is found.
  	 * @param	pt	the <code>Point</code> to be tested.
  	 * @return	<code>true</code> if a charged fluid is found.
  	 */
	private boolean _computeBooleanArray(Point pt)
	{
		Vector indexVector = new Vector(2, 1);
		booleanArray = new boolean[_imageHeight][_imageWidth];
		for (int index = 0; index < _chargedFluidVector.size(); index++) {
		  	ChargedFluid currentCF = (ChargedFluid)_chargedFluidVector.get(index);
		  	if (currentCF.polygonCF.contains(pt)) {
		  		indexVector.add(new Point(index,index));
		  	}
		}
		
		if (indexVector.size() > 0) {
			Point firstIndex = (Point)indexVector.get(0);
			int targetIndex = firstIndex.x;
			ChargedFluid targetCF = (ChargedFluid)_chargedFluidVector.get(firstIndex.x);
			Polygon polygon = targetCF.polygonCF;
			Rectangle rectBound = polygon.getBounds();
			for (int index = 1; index < indexVector.size(); index++) {
				Point indexPt = (Point)indexVector.get(index);
				ChargedFluid currentCF = (ChargedFluid)_chargedFluidVector.get(indexPt.x);
				Polygon currentPolygon = currentCF.polygonCF;
				Rectangle currentBound = currentPolygon.getBounds();
				if (currentBound.width < rectBound.width) {
					targetCF = currentCF;
					polygon = currentPolygon;
					rectBound = currentBound;
					targetIndex = index;
				}
			}
			indexVector.clear();
			for (int i = 0; i < _chargedFluidVector.size(); i++) {
				if (i == targetIndex) {
					continue;
				}
			  	ChargedFluid currentCF = (ChargedFluid)_chargedFluidVector.get(i);
			  	if (polygon.contains(currentCF.polygonCF.getBounds())) {
			  		indexVector.add(new Point(i,i));
			  	}
			}
			for (int y = 0; y < _imageHeight; y++) {
				for (int x = 0; x < _imageWidth; x++) {
					if (polygon.contains(x, y)) {
	        			booleanArray[y][x] = true;
					}
				}
			}	

	      	Iterator itc = targetCF.linkedList.iterator();
	      	while (itc.hasNext()) {
	        	Charge currentCharge = (Charge)itc.next();
	        	Point chargePt = currentCharge.position;
	        	booleanArray[chargePt.y][chargePt.x] = true;
	        }	        
	        
			if (indexVector.size() > 0) {
				for (int j = 0; j < indexVector.size(); j++) {
					Point indexPt = (Point)indexVector.get(j);
					ChargedFluid innerCF = (ChargedFluid)_chargedFluidVector.get(indexPt.x);
					Polygon innerPolygon = innerCF.polygonCF;
			        for (int y = 0; y < _imageHeight; y++) {
						for (int x = 0; x < _imageWidth; x++) {
							if (innerPolygon.contains(x, y)) {
			        			booleanArray[y][x] = false;
							}
						}
					}
					Iterator it = innerCF.linkedList.iterator();
			      	while (it.hasNext()) {
			        	Charge currentCharge = (Charge)it.next();
			        	Point chargePt = currentCharge.position;
			        	booleanArray[chargePt.y][chargePt.x] = false;
			        }		    
				}
			}
			
			return true;
		}
		else {
			return false;
		}	
	}
	
	
	/**
	 * Calculates the smooth image gradient with the specified Gaussian filtering method.
	 */
	private void _computeGaussianFilterArray()
	{	
		double[][] gaussianImageArray = new double[_imageHeight][_imageWidth];
		int xm1, ym1, xp1, yp1;
		int gaussianFilter1[][] = {{0,1,0},{1,4,1},{0,1,0}};
		int gaussianFilter2[][] = {{2,4,5,4,2},{4,9,12,9,4},{5,12,15,12,5},
									{4,9,12,9,4},{2,4,5,4,2}};
		int gaussianFilter3[][] = {{1,4,7,10,7,4,1},{4,12,26,33,26,12,4},{7,26,55,71,55,26,7},
			{10,33,71,91,71,33,10},{7,26,55,71,55,26,7},{4,12,26,33,26,12,4},{1,4,7,10,7,4,1}};
		for (int y = 0; y < _imageHeight; y++) {
			for (int x = 0; x < _imageWidth; x++) {
				gaussianImageArray[y][x] = _imageIntensityArray[y][x];
			}
		}
		int filterLength = 0;
		if (_indexOfFilter > 0) {
		    int gaussianFilter[][] = gaussianFilter1;
		    if (_indexOfFilter == 2) {
		    	gaussianFilter = gaussianFilter2;
		    }
		    else if (_indexOfFilter == 3) {
		    	gaussianFilter = gaussianFilter3;
		    }
		    int filterHeight = gaussianFilter.length;
		    int filterWidth = gaussianFilter[0].length;
		    filterLength = filterHeight / 2;
		    for (int y = filterLength; y < _imageHeight - filterLength; y++) {
		    	for (int x = filterLength; x < _imageWidth - filterLength; x++) {
		    		int tempSum = 0;
		    		for (int j = 0; j < filterHeight; j++) {
		        		for (int i = 0; i < filterWidth; i++) {	
		        			tempSum += gaussianFilter[j][i]*_imageIntensityArray[y-filterLength+j][x-filterLength+i];
		        		}
		        	}
		        	gaussianImageArray[y][x] = tempSum;		
		    	}
		    }
		}
		for (int y = filterLength + 1; y < _imageHeight - filterLength - 1; y++) {
			for (int x = filterLength + 1; x < _imageWidth - filterLength - 1; x++) {
			  	xm1 = x - 1;
		      	ym1 = y - 1;
		      	xp1 = x + 1;
		      	yp1 = y + 1;
			  	_gradientArrayInX[y][x] = (gaussianImageArray[y][xp1] - gaussianImageArray[y][xm1]) / 2.0;
		      	_gradientArrayInY[y][x] = (gaussianImageArray[yp1][x] - gaussianImageArray[ym1][x]) / 2.0;
		    }
		}
		for (int y = 0; y < _imageHeight; y++) {
			for (int x = 0; x < _imageWidth; x++) {			
		      	_imageGradientArray[y][x] = Math.sqrt(_gradientArrayInX[y][x] * _gradientArrayInX[y][x]
		                            		+ _gradientArrayInY[y][x] * _gradientArrayInY[y][x]);	                            		
		    }
		}
		double maxModulus = 0.0;
		for (int y = 0; y < _imageHeight; y++) {
		    for (int x = 0; x < _imageWidth; x++) {
		      	if (_imageGradientArray[y][x] > maxModulus) {
		        	maxModulus = _imageGradientArray[y][x];
		      	}
		    }
		}
		if (maxModulus > 0.0) {
		    for (int y = 0; y < _imageHeight; y++) {
		        for (int x = 0; x < _imageWidth; x++) {
		          	_imageGradientArray[y][x] /= maxModulus;
		          	_gradientArrayInX[y][x] /= maxModulus;
		          	_gradientArrayInY[y][x] /= maxModulus;
		        }
		    }
	 	}
		if (_indexOfPower == 2) {
		    for (int y = 0; y < _imageHeight; y++) {
		        for (int x = 0; x < _imageWidth; x++) {
		          	_imageGradientArray[y][x] = Math.log(255.0*_imageGradientArray[y][x]+1.0);
		        }
		    }
		    maxModulus = Math.log(256.0);
		    for (int y = 0; y < _imageHeight; y++) {
		        for (int x = 0; x < _imageWidth; x++) {
		          	_imageGradientArray[y][x] /= maxModulus;
		        }
		    }
		}
		for (int y = 0; y < _imageHeight; y++) {
			for (int x = 0; x < _imageWidth; x++) {
				_reservedGradientArray[y][x] = _imageGradientArray[y][x];
			}
		}
	}
  	

	/**
	 * Adds the specified seed point to the vector <tt>_seedPoint</tt>.
	 * @param	seedPoint	a seed point to be added.
	 */
	public void addSeedPoint(Point seedPoint)
	{
		_seedPoint.add(seedPoint);
	}


	/**
	 * Clears the seed point vector <tt>_seedPoint</tt>.
	 */
	public void clearSeedPoint()
	{
		_seedPoint.clear();
	}
	
	
	/**
	 * Returns the original image intensity array.
	 * @return	the image intensity array.
	 */
	public double[][] getImageIntensityArray()
	{
		return _imageIntensityArray;
	}
	
	
	/**
	 * Returns the image gradient array subject to the saturation setting.
	 * @return	the saturated image gradient array.
	 */
	public double[][] getImageGradientArray()
	{
		return _imageGradientArray;
	}
	
	
	/**
	 * Returns the reserved (original) image gradient array.
	 * @return	the original image gradient array.
	 */
	public double[][] getReservedGradientArray()
	{
		return _reservedGradientArray;
	}
  
  
	/**
	 * Returns the width of the original image.
	 * @return	the image width.
	 */
	public int getImageWidth()
	{
		return _imageWidth;
	}
	
	
	/**
	 * Returns the height of the original image.
	 * @return	the image height.
	 */
	public int getImageHeight()
	{
		return _imageHeight;
	}
	
	
	/**
	 * Returns the width of the initial box for the charged fluid.
	 * @return	the initial box width.
	 */
	public int getInitialBoxWidth()
	{
		return _initialBoxWidth;
	}
  
  
	/**
	 * Returns the height of the initial box for the charged fluid.
	 * @return	the initial box height.
	 */
	public int getInitialBoxHeight()
	{
		return _initialBoxHeight;
	}
	
	
	/**
	 * Returns the index of the filter being used.
	 * @return	the filter index.
	 */
	public int getIndexOfFilter()
	{
		return _indexOfFilter;
	}
	
	
	/**
	 * Returns the seed point vector <tt>_seedPoint</tt>.
	 * @return	the seed point vector.
	 */
	public Vector getSeedPoint()
	{
		return _seedPoint;	
	}
	
	
	/**
	 * Returns the weight for balancing the potentials, beta.
	 * @return the value of beta.
	 */
	public double getBeta()
	{
		return _beta;	
	}
  
  
	/**
	 * Returns the electrostatic equilibrium condition.
	 * @return	the value of the condition.
	 */
	public double getGamma()
	{
		return _gamma;	
	}
	
	
	/**
	 * Returns the charged fluid vector, <tt>_chargedFluidVector</tt>.
	 * @return	<tt>_chargedFluidVector</tt>.
	 */
	public Vector getChargedFluidVector()
	{
		return _chargedFluidVector;
	}
  
  
	/**
	 * Returns the poisson vector, <tt>_poissonVector</tt>.
	 * @return	<tt>_poissonVector</tt>.
	 */
	public Vector getPoissonVector()
	{
		return _poissonVector;
	}


	/**
	 * Returns the value of the pause index, <tt>_pauseInteger</tt>.
	 * @return	the pause index, <tt>_pauseInteger</tt>.
	 */
	public int getPause()
	{ 
		return _pauseInteger;
	}
	
	
	/**
	 * Sets the width of the image.
	 * @param	imageWidth	the image width.
	 */
	public void setImageWidth(int imageWidth)
	{
		_imageWidth = imageWidth;
	}
	
	
	/**
	 * Sets the height of the image.
	 * @param	imageHeight	the image height.
	 */
	public void setImageHeight(int imageHeight)
	{
		_imageHeight = imageHeight;
	}
  
  
	/**
	 * Sets the index of the filter to be used.
	 * @param	indexOfFilter	the filter index.
	 */
	public void setIndexOfFilter(int indexOfFilter)
	{
		_indexOfFilter = indexOfFilter;
	}
	
	
	/**
	 * Sets the seed point vector with the specified vector.
	 * @param	seedPoint	the seed point vector.
	 */
	public void setSeedPoint(Vector seedPoint)
	{
		_seedPoint = seedPoint;	
	}
  
	
	/**
	 * Sets the weight for balancing the potentials, beta.
	 * @param	beta	the value of beta to be set.
	 */
	public void setBeta(double beta)
	{
		_beta = beta;	
	}

	
	/**
	 * Sets the <tt>_isBreak</tt> property, which decides whether the evolution process 
	 * is terminated by users or not.
	 * @param	isBreak		<code>true</code>, if the evolution process is terminated.
	 */
	public void setBreak(boolean isBreak)
	{
		_isBreak = isBreak;
	}
	
	
	/**
	 * Determines whether the evolution process is terminated by users.
	 * @return	<code>true</code> if the evolution process is terminated.
	 */
	public boolean isBreak()
	{
		return _isBreak;	
	}
	
	
	/**
	 * Sets the evolution process paused by adding the value of <tt>_pauseInteger</tt>.
	 */
	public void setPause()
	{
		_pauseInteger++;
	}
  
  
  	/**
	 * Sets the <tt>_isSquare</tt> property, which decides whether the initial charged 
	 * fluid contour is square or not.
	 * @param	isSquare	<code>true</code> if the initial contour is square.
	 */
  	public void setSquare(boolean isSquare)
  	{
  		_isSquare = isSquare;
  	}
  	
  	
  	/**
	 * Determines whether the initial charged fluid contour is square or not.
	 * @return	<code>true</code> if the initial contour is square.
	 */
  	public boolean isSquare()
  	{
  		return _isSquare;
  	}
}

